import 'dart:async';

import 'package:dio/dio.dart';
import 'package:localstore/localstore.dart';

import 'bagel_db_functions.dart';

String authEndpoint = "https://auth.bageldb.com/api/public";

class Collcetions {
  static String user = "user";
  static String access = "access";
}

class BagelUser {
  String? userID, email, phone;
  BagelUser({
    this.userID,
    this.email,
    this.phone,
  });
}

class AuthEvent {
  BagelUser? user;
  bool loggedin;
  AuthEvent({this.user, this.loggedin = false});
}

class BagelUsersRequest {
  BagelDB instance;
  late Dio dio;
  final _db = Localstore.instance;
  BagelUser? _user;
  BagelUsersRequest(this.instance) {
    dio = Dio(BaseOptions(headers: {
      'Authorization': 'Bearer ${this.instance.token}',
      "Accept-Version": "v1"
    }));
  }

  /// use the getter [user] to get BagelUser or Null to determine if a user is logged in
  @deprecated
  Future<bool> isLoggedIn() async {
    return user != null;
  }

  /// use the getter [user] instead to get the current user;
  @deprecated
  BagelUser? getUser() {
    return user;
  }

  /// use the getter [user] instead to get the currenly logged in BagelUser
  @deprecated
  String? getUserID() {
    return user?.userID;
  }

  Future<bool> _bagelUserActive() async {
    Map<String, dynamic>? access =
        await _db.collection(Collcetions.access).doc("access").get();
    if (access != null) {
      DateTime now = DateTime.now();
      DateTime expiryTime = DateTime.parse(access["expiryTime"]);
      if (now.isAfter(expiryTime)) {
        try {
          await _refresh();
        } catch (err) {
          return false;
        }
      }
      return true;
    } else {
      return false;
    }
  }

  /// Create a user with a an email and password
  Future<BagelUser> create(email, password) async {
    String url = '$authEndpoint/user';
    Map body = {"email": email, "password": password};
    try {
      Response res = await dio.post(url, data: body);
      _storeTokens(res.data);
      return _login(email: email, userID: res.data["user_id"]);
    } catch (err) {
      throw (err);
    }
  }

  Future<String> _getOtpRequestNonce() async {
    Map<String, dynamic>? otpRequest =
        await _db.collection(Collcetions.access).doc("nonce").get();
    if (otpRequest != null) {
      DateTime expiryTime = DateTime.parse(otpRequest["expiryTime"]);
      if (new DateTime.now().isAfter(expiryTime))
        throw ("OTP request has expired, try again");
      return otpRequest["nonce"];
    } else {
      return "";
    }
  }

  _storeOtpRequestNonce(Map otpRequest) {
    DateTime expires =
        new DateTime.now().add(Duration(seconds: otpRequest["expires_in"]));
    _db
        .collection(Collcetions.access)
        .doc("nonce")
        .set({"nonce": otpRequest["nonce"], "expiryTime": expires.toString()});
  }

  /// Request the user receive a One Time Password either via SMS or Email
  Future<String> requestOtp(emailOrPhone) async {
    String url = '$authEndpoint/user/otp';
    Map body = {"emailOrPhone": emailOrPhone};

    Response res = await dio.post(url, data: body);
    _storeOtpRequestNonce(res.data);
    return res.data["nonce"];
  }

  /// Validate the user's received One Time Password
  Future<BagelUser> validateOtp(String otp) async {
    String nonce = await _getOtpRequestNonce();
    String url = '$authEndpoint/user/otp/verify/$nonce';
    Map body = {"otp": otp};
    Response res = await dio.post(url, data: body);
    if (res.statusCode == 200 || res.statusCode == 201) {
      _storeTokens(res.data);
      return _login(userID: res.data["user_id"]);
    } else {
      throw ("wrong authorization code");
    }
  }

  /// Authenticate the user with an email and password
  Future<BagelUser> validate(email, password) async {
    String url = '$authEndpoint/user/verify';
    Map body = {"email": email, "password": password};

    Response res = await dio.post(url, data: body);
    if (res.statusCode == 200 || res.statusCode == 201) {
      _storeTokens(res.data);
      BagelUser u = _login(email: email, userID: res.data["user_id"]);
      return u;
    } else {
      throw ("wrong authorization codes");
    }
  }

  /// Get the currently logged in user's info
  void _getUser() async {
    if (await _bagelUserActive()) {
      String? accessToken = await getAccessToken();
      Options options = Options(headers: {
        'Authorization': 'Bearer $accessToken',
        "Accept-Version": "v1"
      });
      String url = '$authEndpoint/user';
      Response res = await dio.get(url, options: options);
      _login(
          email: res.data["email"],
          phone: res.data["phone"],
          userID: res.data["userID"]);
    }
  }

  _storeBagelUser(BagelUser user) {
    _db
        .collection(Collcetions.user)
        .doc("user")
        .set({"userID": user.userID, "email": user.email, "phone": user.phone});
  }

  Future _storeTokens(data) async {
    String accessToken = data["access_token"],
        refreshToken = data["refresh_token"];
    int expires = data["expires_in"];
    await _db.collection(Collcetions.access).doc("access").set({
      "expiryTime": DateTime.now().add(Duration(seconds: expires)).toString(),
      "accessToken": accessToken,
      "refreshToken": refreshToken
    });
  }

  // ignore: close_sinks
  StreamController<AuthEvent>? _userStateController;

  BagelUser? get user => _user;

  Stream<AuthEvent> get authStateChange {
    void startListen() async {
      bool loggedIn = await _bagelUserActive();
      if (loggedIn) {
        _getUser();
      } else {
        _logout();
      }
    }

    void stopListen() {
      _userStateController?.close();
    }

    _userStateController = StreamController<AuthEvent>.broadcast(
      onListen: startListen,
      onCancel: stopListen,
      sync: true,
    );

    return _userStateController!.stream;
  }

  Future<String?> _getRefreshToken() async {
    Map<String, dynamic>? refreshToken =
        await _db.collection(Collcetions.access).doc("access").get();
    return refreshToken?["refreshToken"];
  }

  Future<String?> getAccessToken() async {
    Map<String, dynamic>? accessToken =
        await _db.collection(Collcetions.access).doc("access").get();
    return accessToken?["accessToken"];
  }

  BagelUser _login({userID, email, phone}) {
    _user = BagelUser();

    if (userID != null) _user!.userID = userID;
    if (email != null) _user!.email = email;
    if (phone != null) _user!.phone = phone;
    _storeBagelUser(_user!);
    AuthEvent event = AuthEvent(user: _user, loggedin: true);
    _userStateController?.add(event);
    return _user!;
  }

  void _logout() async {
    _user = null;
    _userStateController?.add(AuthEvent());
    await _db.collection(Collcetions.access).doc("access").delete();
    await _db.collection(Collcetions.user).doc("userID").delete();
  }

  /// Log the user out
  void logout() => _logout();

  /// is used to refresh the users auth token without having to login with username and password
  Future<Response> _refresh() async {
    if (await _getRefreshToken() == null) {
      throw ("Couldn't get refresh token");
    }

    String url = '$authEndpoint/user/token';
    String? refreshToken = await _getRefreshToken();
    String body =
        'grant_type=refresh_token&refresh_token=$refreshToken&client_id=project-client';
    try {
      Options options = Options(headers: {
        'Content-Type': 'application/x-www-form-urlencoded',
        'Authorization': 'Bearer ${this.instance.token}',
        // "Accept-Version": "v1"
      });
      Response res = await dio.post(url, data: body, options: options);
      if (res.statusCode == 200) {
        await _storeTokens(res.data);
      }
      return res;
    } catch (err) {
      throw (err);
    }
  }
}
